/*
簡易HTMLパーサ
XMLとして破綻しているHTMLをXMLパーサを通せる程度に構造化する。
入力はUTF-8を想定
OCaml 4.02
2015.02.11 - 2015.04.04 Takeshi Yoneki
2020.07.05 ロイター向けに改変
2021.02.27 東京新聞のバグ対処
2021.08.05 テレビ王国の間違った閉じタグ
2021.08.10 DOCTYPEを出力できるようにするため、Document型を用意
2021.08.12 整形出力機能
2021.08.31 終了タグ判定関連の改変
2021.09.05 シングルクオート対応など
2022.01.23 Rustへ移植
2022.12.27 Goへ移植
ostra@asahi.email.ne.jp
*/

package HtmlParser

import (
	"strings"
)

type NodeType int

const (
	NT_Empty NodeType = iota
	NT_Element
	NT_PCData
	NT_CData
	NT_Comment
	NT_DocType
	NT_Decl
	NT_Document
)

type Node interface {
	GetType() NodeType
}

type Attr struct {
	Name  string
	Value string
}

type Empty struct {
}

func (n *Empty) GetType() NodeType {
	return NT_Empty
}

func CreateEmpty() *Empty {
	return &Empty{}
}

type Element struct {
	Tag        string
	Attributes []Attr
	Children   []Node
}

func (n *Element) GetType() NodeType {
	return NT_Element
}

func CreateElement(tag string, attributes []Attr, children []Node) *Element {
	return &Element{tag, attributes, children}
}

func CreateElem(tag string) *Element {
	return &Element{tag, []Attr{}, []Node{}}
}

func AsElement(node Node) (*Element, bool) {
	if node != nil && node.GetType() == NT_Element {
		return node.(*Element), true
	}
	return nil, false
}

func (n *Element) String() string {
	return ToXmlString(n)
}

type PCData struct {
	Data string
}

func (n *PCData) GetType() NodeType {
	return NT_PCData
}

func CreatePCData(data string) *PCData {
	return &PCData{data}
}

type CData struct {
	Data string
}

func (n *CData) GetType() NodeType {
	return NT_CData
}

func CreateCData(data string) *CData {
	return &CData{data}
}

type Comment struct {
	Data string
}

func (n *Comment) GetType() NodeType {
	return NT_Comment
}

func CreateComment(data string) *Comment {
	return &Comment{data}
}

type DocType struct {
	Data string
}

func (n *DocType) GetType() NodeType {
	return NT_DocType
}

func CreateDocType(data string) *DocType {
	return &DocType{data}
}

type Decl struct {
	Data string
}

func (n *Decl) GetType() NodeType {
	return NT_Decl
}

func CreateDecl(data string) *Decl {
	return &Decl{data}
}

type Document struct {
	Children []Node
}

func (n *Document) GetType() NodeType {
	return NT_Document
}

func CreateDocument(children []Node) *Document {
	return &Document{children}
}

func AsDocument(node Node) (*Document, bool) {
	if node != nil && node.GetType() == NT_Document {
		return node.(*Document), true
	}
	return nil, false
}

func (n *Document) String() string {
	return ToXmlString(n)
}

type partKind int

const (
	pk_StartTag partKind = iota
	pk_EndTag
	pk_EmptyTag
	pk_Text
	pk_CData
	pk_Comment
	pk_DocType
	pk_Decl
)

func isSpace(ch rune) bool {
	return ch == ' ' || ch == '\t' || ch == '\n' || ch == '\r'
}

type pp1Item struct {
	k   partKind
	row string
}

// func (item pp1Item) String() string {
// 	return fmt.Sprintf("(%d) %s", int(item.k), item.row)
// }

type pp1Data struct {
	part          []rune
	pool          []pp1Item
	startLessThan bool
}

func (r *pp1Data) addPartToPool(row string) {
	s := strings.ToLower(row)
	var k partKind
	if r.part[0] == '<' && r.part[len(r.part)-1] == '>' {
		// ここに来る場合、文字列長は必ず2以上なのでpart[1]は存在する。
		if r.part[1] == '/' {
			k = pk_EndTag
		} else if r.part[1] == '!' {
			if strings.HasPrefix(s, "<!doctype") {
				k = pk_DocType
			} else if strings.HasPrefix(s, "<![cdata[") && strings.HasSuffix(s, "]]>") {
				k = pk_CData
			} else if strings.HasPrefix(s, "<!--") && strings.HasSuffix(s, "-->") {
				k = pk_Comment
			} else {
				k = pk_StartTag
			}
		} else if r.part[len(r.part)-2] == '/' {
			// ここに来る場合、キャラクタ数は必ず2以上なのでpart[part.len() - 2]は存在する。
			k = pk_EmptyTag
		} else if r.part[1] == '?' {
			if r.part[len(r.part)-2] == '?' {
				k = pk_Decl
			} else {
				k = pk_StartTag
			}
		} else {
			k = pk_StartTag
		}
	} else {
		k = pk_Text
	}
	r.pool = append(r.pool, pp1Item{k, row})
	r.part = []rune{}
}

func (r *pp1Data) parsePhase1Proc(data string) {
	inTag := false
	inQuote := '\x00'
	inAttrValue := false
	inComment := false
	inCData := false
	inScript := false
	r.startLessThan = false

	for _, ch := range data {
		if ch == '\r' {
		} else if inCData {
			r.part = append(r.part, ch)
			if ch == '>' {
				p := string(r.part)
				if strings.HasSuffix(p, "]]>") {
					r.addPartToPool(p)
					inCData = false
					inTag = false
					inAttrValue = false
				}
			}
		} else if inComment {
			r.part = append(r.part, ch)
			if ch == '>' {
				p := string(r.part)
				if strings.HasSuffix(p, "-->") {
					r.addPartToPool(p)
					inComment = false
					inTag = false
					inAttrValue = false
				}
			}
		} else if inScript {
			if ch == '<' {
				if len(r.part) > 0 {
					r.pool = append(r.pool, pp1Item{pk_Text, string(r.part)})
					r.part = []rune{}
				}
				r.startLessThan = true
			}
			r.part = append(r.part, ch)
			if r.startLessThan && (ch == '>' || ch == '-' || ch == '[') {
				p := string(r.part)
				q := strings.ToLower(p)
				if ch == '>' && q == "</script>" {
					r.addPartToPool(p)
					inScript = false
					inTag = false
					inAttrValue = false
				} else if ch == '-' && q == "<!--" {
					inComment = true
				} else if ch == '[' && q == "<![cdata[" {
					inCData = true
				}
			}
		} else if inTag {
			if inQuote != '\x00' {
				if ch == inQuote {
					inQuote = '\x00'
					inAttrValue = false
				}
				r.part = append(r.part, ch)
			} else if ch == '>' {
				r.part = append(r.part, ch)
				p := string(r.part)
				q := strings.ToLower(p)
				// part[0] = '<'ならばキャラクタ数は必ず2以上なのでpart[part.len() - 2]は存在する。
				if r.part[0] == '<' && r.part[len(r.part)-2] != '/' {
					if strings.HasPrefix(q, "<script ") || q == "<script>" {
						inScript = true
					}
				}
				r.addPartToPool(p)
				inTag = false
				inAttrValue = false
			} else if inAttrValue && isSpace(ch) {
				// 「=」後の空白を詰める。属性値のとき空白前の文字が「=」でなければ属性値終了。
				// タグの中なら必ずpart[0]は存在する。
				if r.part[len(r.part)-1] == '=' {
				} else {
					inAttrValue = false
					r.part = append(r.part, ch)
				}
			} else if ch == '=' {
				inAttrValue = true
				r.part = append(r.part, ch)
			} else if ch == '"' || ch == '\'' {
				// 東京新聞で間違ったダブルクオートが混入している。
				// タグの中でのクオート終了直後のクオート開始は無視する。
				// 2021.02.27
				if !inAttrValue {
				} else {
					inQuote = ch
					r.part = append(r.part, ch)
				}
			} else {
				r.part = append(r.part, ch)
			}
			//
			if ch == '-' {
				p := string(r.part)
				if p == "<!--" {
					inComment = true
				}
			} else if ch == '[' {
				p := string(r.part)
				q := strings.ToLower(p)
				if q == "<![cdata[" {
					inCData = true
				}
			}
		} else {
			if ch == '<' {
				if len(r.part) > 0 {
					p := string(r.part)
					r.addPartToPool(p)
				}
				inTag = true
			}
			r.part = append(r.part, ch)
		}
	}
	if len(r.part) > 0 {
		p := string(r.part)
		r.addPartToPool(p)
	}
}

func parsePhase1(data string) []pp1Item {
	p := pp1Data{}
	p.parsePhase1Proc(data)
	return p.pool
}

func FixCharEntity(tx string) string {
	t := tx
	ac := []string{}
	for {
		n, p := FindCharEntity(t)
		if n < 0 {
			ac = append(ac, t)
			break
		} else {
			ac = append(ac, t[0:n])
			ac = append(ac, string(DecodeCharEntity(p)))
			t = t[n+len(p):]
		}
	}
	return strings.Join(ac, "")
}

type psItem struct {
	name  string
	value string
}

type psTagData struct {
	part    []rune
	pool    []psItem
	name    string
	inValue bool
}

func (r *psTagData) enter(force bool) {
	if len(r.part) > 0 || force {
		s := string(r.part)
		r.part = []rune{}
		if r.inValue {
			r.pool = append(r.pool, psItem{r.name, s})
			r.name = ""
		} else {
			if r.name != "" {
				r.pool = append(r.pool, psItem{r.name, r.name})
			}
			r.name = s
		}
	}
}

func (r *psTagData) phaseStartTagProc(k partKind, row string) Node {
	inQuote := '\x00'
	var sub string
	if k == pk_EmptyTag {
		sub = row[1 : len(row)-2]
	} else {
		sub = row[1 : len(row)-1]
	}
	for _, ch := range sub {
		if inQuote != '\x00' {
			if ch == inQuote {
				r.enter(true)
				inQuote = '\x00'
				r.inValue = false
			} else {
				r.part = append(r.part, ch)
			}
		} else {
			if ch == '"' || ch == '\'' {
				r.enter(false)
				inQuote = ch
			} else if ch == '=' {
				r.enter(false)
				r.inValue = true
			} else if isSpace(ch) {
				r.enter(false)
				r.inValue = false
			} else {
				r.part = append(r.part, ch)
			}
		}
	}
	r.enter(false)
	if r.name != "" {
		r.pool = append(r.pool, psItem{r.name, r.name})
	}
	if len(r.pool) > 0 {
		tag := r.pool[0].name
		r.pool = r.pool[1:]
		attr := []Attr{}
		for _, item := range r.pool {
			attr = append(attr, Attr{item.name, FixCharEntity(item.value)})
		}
		return CreateElement(tag, attr, []Node{})
	} else {
		return &Empty{}
	}
}

func fixComment(t string) string {
	// コメント内に"--"を記載してしまうページがある
	return strings.Replace(t[4:len(t)-3], "--", "&#x2d;&#x2d;", -1)
}

type pp2Item struct {
	k partKind
	p Node
}

// func (item pp2Item) String() string {
// 	ss := []string{}
// 	ss = append(ss, fmt.Sprintf("(%d) ", int(item.k)))
// 	if item.p.GetType() == NT_Element {
// 		elem := item.p.(*Element)
// 		ss = append(ss, elem.String())
// 	}
// 	return strings.Join(ss, "")
// }

func parsePhase2(data []pp1Item) []pp2Item {
	result := []pp2Item{}
	for _, item := range data {
		var node Node
		switch item.k {
		case pk_StartTag, pk_EmptyTag:
			st := psTagData{}
			node = st.phaseStartTagProc(item.k, item.row)
		case pk_EndTag:
			tag := item.row[2 : len(item.row)-1]
			node = CreateElement(tag, []Attr{}, []Node{})
		case pk_Text:
			node = CreatePCData(FixCharEntity(item.row))
		case pk_CData:
			node = CreateCData(item.row)
		case pk_Comment:
			node = CreateComment(fixComment(item.row))
		case pk_DocType:
			node = CreateDocType(item.row)
		case pk_Decl:
			node = CreateDecl(item.row)
		}
		result = append(result, pp2Item{item.k, node})
	}
	return result
}

type pp3Data struct {
	endTags map[string]struct{} // GoにはSetがない
	stack   []string
}

func newPp3Data(data []pp2Item) pp3Data {
	s := map[string]struct{}{}
	for _, item := range data {
		if item.k == pk_EndTag {
			elem, ok := AsElement(item.p)
			if ok {
				s[strings.ToLower(elem.Tag)] = struct{}{}
			} else {
				panic("ERROR_BUG")
			}
		}
	}
	return pp3Data{s, []string{}}
}

func (r *pp3Data) hasEnd1(tn string) bool {
	_, ok := r.endTags[tn]
	return ok
}

func (r *pp3Data) stackContains(n string) bool {
	for _, s := range r.stack {
		if s == n {
			return true
		}
	}
	return false
}

func mkInnerData(tn string, data []pp2Item) ([]pp2Item, []pp2Item, bool) {
	bracket := 1
	inner := []pp2Item{}
	for bracket > 0 && len(data) > 0 {
		d := data[0]
		data = data[1:]
		elem, ok := AsElement(d.p)
		if ok {
			tag := elem.Tag
			if strings.ToLower(tag) == tn {
				switch d.k {
				case pk_StartTag:
					bracket++
				case pk_EndTag:
					bracket--
				}
			}
		}
		inner = append(inner, d)
	}
	return data, inner, bracket == 0
}

func (r *pp3Data) mkTree(dx []pp2Item) ([]pp2Item, []Node, string) {
	breakFlag := false
	sibling := []Node{}
	q := ""
	for len(dx) > 0 && !breakFlag {
		d := dx[0]
		dx = dx[1:]
		switch d.k {
		case pk_StartTag:
			e, ok := AsElement(d.p)
			if ok {
				tn := strings.ToLower(e.Tag)
				if r.hasEnd1(tn) && len(dx) > 0 {
					rest1, inner, has_end2 := mkInnerData(tn, dx)
					if has_end2 {
						r.stack = append(r.stack, tn)
						rest2, child, cue := r.mkTree(inner)
						dx = append(rest2, rest1...)
						r.stack = r.stack[:len(r.stack)-1]
						sibling = append(sibling, CreateElement(e.Tag, e.Attributes, child))
						if len(r.stack) > 0 && cue != tn && r.stackContains(cue) {
							q = cue
							breakFlag = true
						}
					} else {
						sibling = append(sibling, d.p)
					}
				} else {
					sibling = append(sibling, d.p)
				}
			}
			// else {
			// 	panic("ERROR_BUG 2")
			// }
		case pk_EndTag:
			e, ok := AsElement(d.p)
			if ok {
				tn := strings.ToLower(e.Tag)
				currentTag := ""
				if len(r.stack) > 0 {
					currentTag = r.stack[len(r.stack)-1]
				}
				// テレビ王国の間違った閉じタグ 2021.08.05
				if tn != currentTag {
					sibling = append(sibling, &Comment{"/" + e.Tag})
				} else {
					q = tn
					breakFlag = true
				}
			} else {
				panic("ERROR_BUG 3")
			}
		default:
			sibling = append(sibling, d.p)
		}
	}
	return dx, sibling, q
}

func parsePhase3(data []pp2Item) []Node {
	p3 := newPp3Data(data)
	_, s, _ := p3.mkTree(data)
	return s
}

func Parse(data string) *Document {
	r1 := parsePhase1(data)
	r2 := parsePhase2(r1)
	r3 := parsePhase3(r2)
	return CreateDocument(r3)
}

func Clone(node Node) Node {
	switch node.GetType() {
	//case NT_Empty:
	//	return CreateEmpty()
	case NT_Element:
		e := node.(*Element)
		tag2 := strings.Clone(e.Tag)
		att2 := []Attr{}
		for _, at := range e.Attributes {
			att2 = append(att2, Attr{strings.Clone(at.Name), strings.Clone(at.Value)})
		}
		chl2 := []Node{}
		for _, ch := range e.Children {
			chl2 = append(chl2, Clone(ch))
		}
		return CreateElement(tag2, att2, chl2)
	case NT_PCData:
		return CreatePCData(strings.Clone(node.(*PCData).Data))
	case NT_CData:
		return CreateCData(strings.Clone(node.(*CData).Data))
	case NT_Comment:
		return CreateComment(strings.Clone(node.(*Comment).Data))
	case NT_DocType:
		return CreateDocType(strings.Clone(node.(*DocType).Data))
	case NT_Decl:
		return CreateDecl(strings.Clone(node.(*Decl).Data))
	case NT_Document:
		d := node.(*Document)
		chl2 := []Node{}
		for _, ch := range d.Children {
			chl2 = append(chl2, Clone(ch))
		}
		return CreateDocument(chl2)
	default:
		return CreateEmpty()
	}
}

func (d *Document) Clone() *Document {
	return Clone(d).(*Document)
}

// ToXmlString

// 実体参照にすべき文字は実体参照にする
func toCharEntity(isAttr bool, t string) string {
	t = strings.ReplaceAll(t, "&", "&amp;")
	t = strings.ReplaceAll(t, ">", "&gt;")
	t = strings.ReplaceAll(t, "<", "&lt;")
	if isAttr {
		// 2023.01.25 属性値の場合に&amp;などを加える。
		t = strings.ReplaceAll(t, "\"", "&quot;")
		t = strings.ReplaceAll(t, "'", "&quos;")
	}
	return t
}

func fixAttrValue(t string) string {
	return toCharEntity(true, t)
}

func fixText(t string) string {
	return toCharEntity(false, t)
}

// 無名関数を変数に入れることでの関数内関数では、再帰できない。
func onNode(n Node, ac []string) []string {
	switch n.GetType() {
	case NT_Element:
		e := n.(*Element)
		ac = append(ac, "<"+e.Tag)
		for _, a := range e.Attributes {
			ac = append(ac, " "+a.Name+"=\""+fixAttrValue(a.Value)+"\"")
		}
		if len(e.Children) == 0 {
			ac = append(ac, " />")
		} else {
			ac = append(ac, ">")
			for _, ch := range e.Children {
				ac = onNode(ch, ac)
			}
			ac = append(ac, "</"+e.Tag+">")
		}
	case NT_Comment:
		e := n.(*Comment)
		ac = append(ac, "<!--"+e.Data+"-->")
	case NT_PCData:
		e := n.(*PCData)
		ac = append(ac, fixText(e.Data))
	case NT_CData:
		e := n.(*CData)
		ac = append(ac, fixText(e.Data))
	case NT_DocType:
		e := n.(*DocType)
		ac = append(ac, e.Data)
	case NT_Decl:
		e := n.(*Decl)
		ac = append(ac, e.Data)
	case NT_Document:
		e := n.(*Document)
		for _, ch := range e.Children {
			ac = onNode(ch, ac)
		}
	}
	return ac
}

func ToXmlString(node Node) string {
	texts := onNode(node, []string{})
	if node.GetType() == NT_Document {
	} else {
		texts = append(texts, "\n")
	}
	return strings.Join(texts, "")
}

func IsElement(node Node) bool {
	return node.GetType() == NT_Element
}

func IsText(node Node) bool {
	t := node.GetType()
	return t == NT_PCData || t == NT_CData
}

func (e *Element) RemoveAttribute(name string) *Element {
	ax := []Attr{}
	for _, a := range e.Attributes {
		if a.Name != name {
			ax = append(ax, a)
		}
	}
	e.Attributes = ax
	return e
}

func RemoveAttribute(node Node, name string) Node {
	e, ok := AsElement(node)
	if ok {
		return e.RemoveAttribute(name)
	}
	return node
}

func SetAttribute(node Node, name string, value string) Node {
	e, ok := AsElement(node)
	if ok {
		return e.SetAttribute(name, value)
	}
	return node
}

func (e *Element) SetAttribute(name string, value string) *Element {
	e.RemoveAttribute(name)
	e.Attributes = append(e.Attributes, Attr{name, value})
	return e
}

func RemoveAttributeAll(node Node) Node {
	e, ok := AsElement(node)
	if ok {
		return e.RemoveAttributeAll()
	}
	return node
}

func (e *Element) RemoveAttributeAll() *Element {
	e.Attributes = []Attr{}
	return e
}

func GetAttribute(node Node, name string) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetAttribute(name)
	}
	return ""
}

func (e *Element) GetAttribute(name string) string {
	for _, a := range e.Attributes {
		if a.Name == name {
			return a.Value
		}
	}
	return ""
}

func GetDocumentElement(node Node) (*Element, bool) {
	switch node.GetType() {
	case NT_Document:
		d, _ := AsDocument(node)
		return d.GetDocumentElement()
	case NT_Element:
		e, _ := AsElement(node)
		return e, true
	}
	return nil, false
}

func (d *Document) GetDocumentElement() (*Element, bool) {
	for _, ch := range d.Children {
		elem, ok := AsElement(ch)
		if ok {
			return elem, true
		}
	}
	return nil, false
}

func GetDataText(node Node) string {
	switch node.GetType() {
	case NT_PCData:
		return node.(*PCData).Data
	case NT_CData:
		return node.(*CData).Data
	}
	return ""
}

func (e *Element) GetElementText() string {
	ts := []string{}
	for _, ch := range e.Children {
		ts = append(ts, GetDataText(ch))
	}
	return strings.Join(ts, "")
}

func GetElementText(node Node) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetElementText()
	}
	return ""
}

func (e *Element) GetTextContent() string {
	return e.GetElementText()
}

func GetTextContent(node Node) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetElementText()
	} else {
		return GetDataText(node)
	}
}

func (e *Element) GetTextContentRecur() string {
	acc := ""
	for _, ch := range e.Children {
		acc += GetTextContentRecur(ch)
	}
	return acc
}

func GetTextContentRecur(node Node) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetTextContentRecur()
	} else {
		return GetDataText(node)
	}
}

func (e *Element) GetElementsByTagName(name string) []*Element {
	r := []*Element{}
	tt := strings.Split(e.Tag, ":")
	if e.Tag == name || (len(tt) > 1 && tt[1] == name) {
		r = append(r, e)
	}
	for _, ch := range e.Children {
		e, ok := AsElement(ch)
		if ok {
			r = append(r, e.GetElementsByTagName(name)...)
		}
	}
	return r
}

func (d *Document) GetElementsByTagName(name string) []*Element {
	r := []*Element{}
	for _, ch := range d.Children {
		e, ok := AsElement(ch)
		if ok {
			r = append(r, e.GetElementsByTagName(name)...)
		}
	}
	return r
}

func GetElementsByTagName(node Node, name string) []*Element {
	r := []*Element{}
	switch node.GetType() {
	case NT_Element:
		e, _ := AsElement(node)
		return e.GetElementsByTagName(name)
	case NT_Document:
		d, _ := AsDocument(node)
		return d.GetElementsByTagName(name)
	}
	return r
}

func GetChildren(node Node) []Node {
	switch node.GetType() {
	case NT_Element:
		e, _ := AsElement(node)
		return e.Children
	case NT_Document:
		d, _ := AsDocument(node)
		return d.Children
	}
	return []Node{}
}

func SetChildren(node Node, children []Node) Node {
	switch node.GetType() {
	case NT_Element:
		e, _ := AsElement(node)
		e.Children = children
	case NT_Document:
		d, _ := AsDocument(node)
		d.Children = children
	}
	return node
}

func (e *Element) AppendChild(ch Node) *Element {
	e.Children = append(e.Children, ch)
	return e
}

func (d *Document) AppendChild(ch Node) *Document {
	d.Children = append(d.Children, ch)
	return d
}

func AppendChild(node Node, ch Node) Node {
	switch node.GetType() {
	case NT_Element:
		e, _ := AsElement(node)
		return e.AppendChild(ch)
	case NT_Document:
		d, _ := AsDocument(node)
		return d.AppendChild(ch)
	}
	return node
}

func (e *Element) GetTagName() string {
	return e.Tag
}

func GetTagName(node Node) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetTagName()
	}
	return ""
}

func (e *Element) GetLocalTagName() string {
	tt := strings.Split(e.Tag, ":")
	if len(tt) > 1 {
		return tt[1]
	}
	return e.Tag
}

func GetLocalTagName(node Node) string {
	e, ok := AsElement(node)
	if ok {
		return e.GetLocalTagName()
	}
	return ""
}

func (e *Element) SetTagName(name string) *Element {
	e.Tag = name
	return e
}

func SetTagName(node Node, name string) Node {
	e, ok := AsElement(node)
	if ok {
		return e.SetTagName(name)
	}
	return node
}

// RuleTransformer

func ruleTransformerSub(trans func(Node, string) Node, node Node, path string) Node {
	p := path
	if IsElement(node) {
		p += ("/" + GetTagName(node))
	}
	r := trans(node, p)
	if r == nil {
		r = CreateEmpty()
	}
	switch r.GetType() {
	case NT_Element, NT_Document:
		ns := []Node{}
		for _, ch := range GetChildren(r) {
			n := ruleTransformerSub(trans, ch, p)
			if n.GetType() != NT_Empty {
				ns = append(ns, n)
			}
		}
		SetChildren(r, ns)
	}
	return r
}

func RuleTransformer(trans func(Node, string) Node, node Node) Node {
	return ruleTransformerSub(trans, node, "")
}

// Pretty

func isSpaceString(s string) bool {
	for _, ch := range s {
		if !isSpace(ch) {
			return false
		}
	}
	return true
}

func hasTextPCData(children []Node) bool {
	for _, ch := range children {
		ok := false
		switch ch.GetType() {
		case NT_PCData:
			ok = !isSpaceString(ch.(*PCData).Data)
		case NT_CData:
			ok = !isSpaceString(ch.(*CData).Data)
		default:
			ok = false
		}
		if ok {
			return true
		}
	}
	return false
}

func deleteSpacePCData(children []Node) []Node {
	if len(children) == 0 || hasTextPCData(children) {
		return children
	}
	r := []Node{}
	for _, ch := range children {
		ok := true
		switch ch.GetType() {
		case NT_PCData:
			ok = !isSpaceString(ch.(*PCData).Data)
		case NT_CData:
			ok = !isSpaceString(ch.(*CData).Data)
		default:
			ok = true
		}
		if ok {
			r = append(r, ch)
		}
	}
	return r
}

func deleteSpace(doc Node) Node {
	doc2 := RuleTransformer(
		func(node Node, path string) Node {
			switch node.GetType() {
			case NT_Element, NT_Document:
				node = SetChildren(node, deleteSpacePCData(GetChildren(node)))
			}
			return node
		}, doc)
	return doc2
}

func calcIndentCount(path string) int {
	count := 0
	for _, c := range path {
		if c == '/' {
			count++
		}
	}
	return count
}

func mkIndentPCData(tab string, count int) Node {
	if count < 1 {
		return CreatePCData("")
	} else {
		acc := []string{}
		for i := 0; i < count; i++ {
			acc = append(acc, tab)
		}
		return CreatePCData(strings.Join(acc, ""))
	}
}

func mkEndLine() Node {
	return CreatePCData("\n")
}

func indentSpace(doc Node, tab string) Node {
	doc2 := RuleTransformer(
		func(node Node, path string) Node {
			switch node.GetType() {
			case NT_Element:
				chldren := GetChildren(node)
				if len(chldren) == 0 || hasTextPCData(chldren) {
				} else {
					indentCount := calcIndentCount(path)
					children2 := []Node{}
					children2 = append(children2, mkEndLine())
					for _, ch := range chldren {
						children2 = append(children2, mkIndentPCData(tab, indentCount))
						children2 = append(children2, ch)
						children2 = append(children2, mkEndLine())
					}
					children2 = append(children2, mkIndentPCData(tab, indentCount-1))
					SetChildren(node, children2)
				}
			case NT_Document:
				chldren := GetChildren(node)
				children2 := []Node{}
				for _, ch := range chldren {
					children2 = append(children2, ch)
					children2 = append(children2, mkEndLine())
				}
				SetChildren(node, children2)
			}
			return node
		}, doc)
	return doc2
}

func Pretty(doc Node, tab string) Node {
	doc2 := deleteSpace(doc)
	doc3 := indentSpace(doc2, tab)
	return doc3
}

// func Test01() {
// 	a := CreateElement("a", []Attr{}, []Node{})
// 	ToXmlString(a)
// }
