/*
Calibreで生成したEPUB2ファイルをEPUB3ファイルに変換するプログラム
2022.12.30 Takeshi Yoneki
ostra@asahi.email.ne.jp

確認環境
go version go1.19.3 windows/amd64 (Windows 10)
go version go1.18.1 linux/amd64 (Ubuntu 22.04)

go.mod
----------
module epub2to3_GO

go 1.19
----------

epub2to3_GOから独自簡易HTMLパーサを利用するように変更。
*/

package main

import (
	"archive/zip"
	"epubfix/HtmlParser"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"strconv"
	"strings"
)

type Config struct {
	vertical bool
	src      string
	dst      string
	css      string
	cssText  string
	id       bool
}

func newConfig() Config {
	return Config{false, "", "", "", "", true}
}

func (cfg *Config) parseParam(args []string) {
	if len(args) < 2 {
		return
	}
	p := args[0]
	v := args[1]
	switch p {
	case "--direction", "-r":
		cfg.vertical = strings.ToLower(v) == "v"
	case "--src", "-s", "--input", "-i":
		cfg.src = v
	case "--dst", "-d", "--output", "-o":
		cfg.dst = v
	case "--css", "-c":
		cfg.css = v
	case "--id", "--kobo", "-k":
		cfg.id = strings.ToLower(v) == "true"
	default:
		cfg.parseParam(args[1:])
	}
	cfg.parseParam(args[2:])
}

func (cfg *Config) loadCss() {
	if cfg.css != "" {
		f, err := os.Open(cfg.css)
		if err != nil {
			fmt.Println(err)
			return
		}
		defer f.Close()
		data, err := io.ReadAll(f)
		if err != nil {
			fmt.Println(err)
			return
		}
		cfg.cssText = string(data)
	}
}

func hasSvgHtml(rd io.Reader) bool {
	data, err := io.ReadAll(rd)
	if err != nil {
		fmt.Println(err)
		return false
	}
	doc := HtmlParser.Parse(string(data))
	return len(doc.GetElementsByTagName("svg")) > 0
}

func getSvgHtmlPathList(inPath string) []string {
	var result []string
	inz, err := zip.OpenReader(inPath)
	if err != nil {
		fmt.Println(err)
		return result
	}
	defer inz.Close()
	for _, ent := range inz.File {
		if ent.FileInfo().IsDir() {
		} else {
			n := strings.ToLower(ent.Name)
			if strings.HasSuffix(n, ".html") || strings.HasSuffix(n, ".xhtml") {
				f, _ := ent.Open()
				defer f.Close()
				if hasSvgHtml(f) {
					result = append(result, ent.Name)
				}
			}
		}
	}
	return result
}

func getNavPathFromNcxPath(path string) string {
	return filepath.Join(filepath.Dir(path), "nav.html")
}

func getNcxPath(elem *HtmlParser.Element) string {
	for _, mE := range elem.GetElementsByTagName("manifest") {
		for _, iE := range mE.GetElementsByTagName("item") {
			if iE.GetAttribute("media-type") == "application/x-dtbncx+xml" {
				return iE.GetAttribute("href")
			}
		}
	}
	return ""
}

func fixOpf(cfg *Config, rd io.Reader, wt io.Writer, svgHtmlList []string) {
	data, err := io.ReadAll(rd)
	if err != nil {
		fmt.Println(err)
		return
	}
	doc := HtmlParser.Parse(string(data))
	rootE, ok := HtmlParser.GetDocumentElement(doc)
	if !ok {
		return
	}
	ncxPath := getNcxPath(rootE)
	d := HtmlParser.RuleTransformer(
		func(node HtmlParser.Node, path string) HtmlParser.Node {
			elem, ok := HtmlParser.AsElement(node)
			if ok {
				switch elem.GetLocalTagName() {
				case "package":
					elem.SetAttribute("version", "3.0")
				case "creator", "contributor":
					elem.RemoveAttributeAll()
				case "identifier":
					elem.RemoveAttribute("scheme")
				case "date":
					t := elem.GetElementText()
					dt := t[0:19] + "Z"
					e := HtmlParser.CreateElem("meta")
					e.SetAttribute("property", "dcterms:modified")
					e.AppendChild(HtmlParser.CreateCData(dt))
					return e
				case "manifest":
					if ncxPath != "" {
						e := HtmlParser.CreateElem("item")
						e.SetAttribute("href", getNavPathFromNcxPath(ncxPath))
						e.SetAttribute("id", "nav")
						e.SetAttribute("media-type", "application/xhtml+xml")
						e.SetAttribute("properties", "nav")
						HtmlParser.AppendChild(node, e)
					}
				case "spine":
					if cfg.vertical {
						elem.SetAttribute("page-progression-direction", "rtl")
					}
				case "item":
					h := elem.GetAttribute("href")
					for _, s := range svgHtmlList {
						if h == s {
							elem.SetAttribute("properties", "svg")
						}
					}
				}
			}
			return node
		}, rootE)
	xml := HtmlParser.ToXmlString(HtmlParser.Pretty(d, "\t"))
	wt.Write([]byte(xml))
}

type NavPoint struct {
	npTitle string
	src     string
}

func convNcxToNav(rd io.Reader, wt io.Writer) {
	data, err := io.ReadAll(rd)
	if err != nil {
		fmt.Println(err)
		return
	}
	doc := HtmlParser.Parse(string(data))
	rootE, ok := HtmlParser.GetDocumentElement(doc)
	if !ok {
		return
	}
	title := ""
	for _, dtE := range rootE.GetElementsByTagName("docTitle") {
		for _, tE := range dtE.GetElementsByTagName("text") {
			title = string(tE.GetElementText())
			break
		}
	}
	npList := []NavPoint{}
	for _, npE := range rootE.GetElementsByTagName("navPoint") {
		npTitle := ""
		for _, nlE := range npE.GetElementsByTagName("navLabel") {
			for _, tE := range nlE.GetElementsByTagName("text") {
				npTitle = string(tE.GetElementText())
				break
			}
		}
		src := ""
		for _, nC := range npE.GetElementsByTagName("content") {
			src = nC.GetAttribute("src")
		}
		if npTitle != "" && src != "" {
			npList = append(npList, NavPoint{npTitle, src})
		}
	}
	tmpText := "<html xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:epub=\"http://www.idpf.org/2007/ops\"><head><meta charset=\"UTF-8\" /><title /></head><body><nav epub:type=\"toc\"><h1 /><ol /></nav></body></html>"
	tmpDoc := HtmlParser.Parse(tmpText)
	d := HtmlParser.RuleTransformer(
		func(node HtmlParser.Node, path string) HtmlParser.Node {
			elem, ok := HtmlParser.AsElement(node)
			if ok {
				switch elem.GetLocalTagName() {
				case "title", "h1":
					elem.AppendChild(HtmlParser.CreateCData(title))
				case "ol":
					for _, np := range npList {
						aE := HtmlParser.CreateElem("a")
						aE.SetAttribute("href", np.src)
						aE.AppendChild(HtmlParser.CreateCData(np.npTitle))
						liE := HtmlParser.CreateElem("li")
						liE.AppendChild(aE)
						elem.AppendChild(liE)
					}
				}
			}
			return node
		}, tmpDoc)
	xml := HtmlParser.ToXmlString(HtmlParser.Pretty(d, "\t"))
	wt.Write([]byte(xml))
}

func fixCss(cfg *Config, rd io.Reader, wt io.Writer) {
	d0, _ := ioutil.ReadAll(rd)
	s0 := string(d0)
	s1 := "\n\n@page { margin: 10%; }\n"
	s2 := "html { -epub-writing-mode: vertical-rl; writing-mode: vertical-rl; }\n"
	s3 := cfg.cssText
	s := ""
	if cfg.vertical {
		s = s0 + s1 + s2 + s3
	} else {
		s = s0 + s1 + s3
	}
	wt.Write([]byte(s))
}

func fixHtml(cfg *Config, rd io.Reader, wt io.Writer) {
	data, err := io.ReadAll(rd)
	if err != nil {
		fmt.Println(err)
		return
	}
	doc := HtmlParser.Parse(string(data))
	rootE, ok := HtmlParser.GetDocumentElement(doc)
	if !ok {
		return
	}
	idNumber := 1
	d := HtmlParser.RuleTransformer(
		func(node HtmlParser.Node, path string) HtmlParser.Node {
			elem, ok := HtmlParser.AsElement(node)
			if ok {
				switch elem.GetLocalTagName() {
				case "meta":
					if elem.GetAttribute("http-equiv") != "" {
						return HtmlParser.CreateEmpty()
					}
				case "link":
					if elem.GetAttribute("rel") == "alternate stylesheet" {
						return HtmlParser.CreateEmpty()
					}
				case "p", "div", "span", "h1", "h2", "a", "img":
					if cfg.id && elem.GetAttribute("id") == "" {
						id := "kobo." + strconv.Itoa(idNumber) + ".1"
						idNumber++
						elem.SetAttribute("id", id)
					}
				}
			}
			return node
		}, rootE)
	xml := HtmlParser.ToXmlString(d)
	//xml := HtmlParser.ToXmlString(HtmlParser.Pretty(d, "\t"))
	wt.Write([]byte(xml))
}

func fixZip(cfg *Config, inPath string, outPath string) {
	svgHtmlList := getSvgHtmlPathList(inPath)
	//fmt.Println(svgHtmlList)
	inz, err := zip.OpenReader(inPath)
	if err != nil {
		fmt.Println(err)
		return
	}
	defer inz.Close()
	outf, err := os.Create(outPath)
	if err != nil {
		fmt.Println(err)
		return
	}
	outz := zip.NewWriter(outf)
	defer outz.Close()
	for _, ent := range inz.File {
		if ent.FileInfo().IsDir() {
		} else {
			n := strings.ToLower(ent.Name)
			if strings.HasSuffix(n, ".opf") {
				rd, _ := ent.Open()
				defer rd.Close()
				wt, _ := outz.Create(ent.Name)
				fixOpf(cfg, rd, wt, svgHtmlList)
			} else if strings.HasSuffix(n, ".ncx") {
				outz.Copy(ent)
				rd, _ := ent.Open()
				defer rd.Close()
				wt, _ := outz.Create(getNavPathFromNcxPath(ent.Name))
				convNcxToNav(rd, wt)
			} else if strings.HasSuffix(n, ".css") {
				rd, _ := ent.Open()
				defer rd.Close()
				wt, _ := outz.Create(ent.Name)
				fixCss(cfg, rd, wt)
			} else if strings.HasSuffix(n, ".html") || strings.HasSuffix(n, ".xhtml") {
				rd, _ := ent.Open()
				defer rd.Close()
				wt, _ := outz.Create(ent.Name)
				fixHtml(cfg, rd, wt)
			} else {
				outz.Copy(ent)
			}
		}
	}
}

func main() {
	cfg := newConfig()
	cfg.parseParam(os.Args[1:])
	cfg.loadCss()
	fixZip(&cfg, cfg.src, cfg.dst)
}
